@@ -16,32 +16,30 @@
  * Boston, MA 021110-1307, USA.
  */
 
+#include "kerncompat.h"
+#include <sys/ioctl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <sys/ioctl.h>
 #include <errno.h>
-#include <sys/stat.h>
 #include <time.h>
-#include <assert.h>
-#include <inttypes.h>
-#include <sys/wait.h>
 #include <getopt.h>
-
-#include "kerncompat.h"
+#include <dirent.h>
+#include <signal.h>
+#include <stdbool.h>
 #include "kernel-shared/ctree.h"
-#include "ioctl.h"
 #include "common/utils.h"
-#include "kernel-shared/volumes.h"
-#include "kernel-shared/disk-io.h"
 #include "common/open-utils.h"
-#include "cmds/commands.h"
 #include "common/help.h"
 #include "common/path-utils.h"
 #include "common/device-utils.h"
+#include "common/string-utils.h"
+#include "common/messages.h"
+#include "cmds/commands.h"
 #include "mkfs/common.h"
+#include "ioctl.h"
 
 static int print_replace_status(int fd, const char *path, int once);
 static char *time2string(char *buf, size_t s, __u64 t);
@@ -114,8 +112,9 @@ static const char *const cmd_replace_start_usage[] = {
 	"       correct checksum. Devices which are currently mounted are",
 	"       never allowed to be used as the <targetdev>",
 	"-B     do not background",
-	"--enqueue    wait if there's another exclusive operation running,",
-	"             otherwise continue",
+	"--enqueue       wait if there's another exclusive operation running,",
+	"                otherwise continue",
+	"-K|--nodiscard  do not perform whole device TRIM",
 	NULL
 };
 
@@ -133,36 +132,41 @@ static int cmd_replace_start(const struct cmd_struct *cmd,
 	char *path;
 	char *srcdev;
 	char *dstdev = NULL;
-	int avoid_reading_from_srcdev = 0;
-	int force_using_targetdev = 0;
+	bool avoid_reading_from_srcdev = false;
+	bool force_using_targetdev = false;
 	u64 dstdev_block_count;
-	int do_not_background = 0;
+	bool do_not_background = false;
 	DIR *dirstream = NULL;
 	u64 srcdev_size;
 	u64 dstdev_size;
 	bool enqueue = false;
+	bool discard = true;
 
 	optind = 0;
 	while (1) {
 		int c;
-		enum { GETOPT_VAL_ENQUEUE = 256 };
+		enum { GETOPT_VAL_ENQUEUE = GETOPT_VAL_FIRST };
 		static const struct option long_options[] = {
 			{ "enqueue", no_argument, NULL, GETOPT_VAL_ENQUEUE},
+			{ "nodiscard", no_argument, NULL, 'K' },
 			{ NULL, 0, NULL, 0}
 		};
 
-		c = getopt_long(argc, argv, "Brf", long_options, NULL);
+		c = getopt_long(argc, argv, "BKrf", long_options, NULL);
 		if (c < 0)
 			break;
 		switch (c) {
 		case 'B':
-			do_not_background = 1;
+			do_not_background = true;
+			break;
+		case 'K':
+			discard = false;
 			break;
 		case 'r':
-			avoid_reading_from_srcdev = 1;
+			avoid_reading_from_srcdev = true;
 			break;
 		case 'f':
-			force_using_targetdev = 1;
+			force_using_targetdev = true;
 			break;
 		case GETOPT_VAL_ENQUEUE:
 			enqueue = true;
@@ -197,14 +201,12 @@ static int cmd_replace_start(const struct cmd_struct *cmd,
 	status_args.result = BTRFS_IOCTL_DEV_REPLACE_RESULT_NO_RESULT;
 	ret = ioctl(fdmnt, BTRFS_IOC_DEV_REPLACE, &status_args);
 	if (ret < 0) {
-		fprintf(stderr,
-			"ERROR: ioctl(DEV_REPLACE_STATUS) failed on \"%s\": %m",
-			path);
+		error("ioctl(DEV_REPLACE_STATUS) failed on \"%s\": %m", path);
 		if (status_args.result != BTRFS_IOCTL_DEV_REPLACE_RESULT_NO_RESULT)
-			fprintf(stderr, ", %s\n",
+			pr_stderr(LOG_DEFAULT, ", %s\n",
 				replace_dev_result2string(status_args.result));
 		else
-			fprintf(stderr, "\n");
+			pr_stderr(LOG_DEFAULT, "\n");
 		goto leave_with_error;
 	}
 
@@ -297,6 +299,7 @@ static int cmd_replace_start(const struct cmd_struct *cmd,
 		BTRFS_DEVICE_PATH_NAME_MAX);
 	ret = btrfs_prepare_device(fddstdev, dstdev, &dstdev_block_count, 0,
 			PREP_DEVICE_ZERO_END | PREP_DEVICE_VERBOSE |
+			(discard ? PREP_DEVICE_DISCARD : 0) |
 			(zoned ? PREP_DEVICE_ZONED : 0));
 	if (ret)
 		goto leave_with_error;
@@ -319,14 +322,12 @@ static int cmd_replace_start(const struct cmd_struct *cmd,
 	ret = ioctl(fdmnt, BTRFS_IOC_DEV_REPLACE, &start_args);
 	if (do_not_background) {
 		if (ret < 0) {
-			fprintf(stderr,
-				"ERROR: ioctl(DEV_REPLACE_START) failed on \"%s\": %m",
-				path);
+			error("ioctl(DEV_REPLACE_START) failed on \"%s\": %m", path);
 			if (start_args.result != BTRFS_IOCTL_DEV_REPLACE_RESULT_NO_RESULT)
-				fprintf(stderr, ", %s\n",
+				pr_stderr(LOG_DEFAULT, ", %s\n",
 					replace_dev_result2string(start_args.result));
 			else
-				fprintf(stderr, "\n");
+				pr_stderr(LOG_DEFAULT, "\n");
 
 			if (errno == EOPNOTSUPP)
 				warning("device replace of RAID5/6 not supported with this kernel");
@@ -423,13 +424,12 @@ static int print_replace_status(int fd, const char *path, int once)
 		args.result = BTRFS_IOCTL_DEV_REPLACE_RESULT_NO_RESULT;
 		ret = ioctl(fd, BTRFS_IOC_DEV_REPLACE, &args);
 		if (ret < 0) {
-			fprintf(stderr, "ERROR: ioctl(DEV_REPLACE_STATUS) failed on \"%s\": %m",
-				path);
+			error("ioctl(DEV_REPLACE_STATUS) failed on \"%s\": %m", path);
 			if (args.result != BTRFS_IOCTL_DEV_REPLACE_RESULT_NO_RESULT)
-				fprintf(stderr, ", %s\n",
+				pr_stderr(LOG_DEFAULT, ", %s\n",
 					replace_dev_result2string(args.result));
 			else
-				fprintf(stderr, "\n");
+				pr_stderr(LOG_DEFAULT, "\n");
 			return ret;
 		}
 
@@ -494,9 +494,8 @@ static int print_replace_status(int fd, const char *path, int once)
 		if (!skip_stats)
 			num_chars += printf(
 				", %llu write errs, %llu uncorr. read errs",
-				(unsigned long long)status->num_write_errors,
-				(unsigned long long)
-				 status->num_uncorrectable_read_errors);
+				status->num_write_errors,
+				status->num_uncorrectable_read_errors);
 		if (once || prevent_loop) {
 			printf("\n");
 			break;
@@ -520,7 +519,7 @@ time2string(char *buf, size_t s, __u64 t)
 	time_t t_time_t;
 
 	t_time_t = (time_t)t;
-	assert((__u64)t_time_t == t);
+	ASSERT((__u64)t_time_t == t);
 	localtime_r(&t_time_t, &t_tm);
 	strftime(buf, s, "%e.%b %T", &t_tm);
 	return buf;
@@ -530,7 +529,7 @@ static char *
 progress2string(char *buf, size_t s, int progress_1000)
 {
 	snprintf(buf, s, "%d.%01d%%", progress_1000 / 10, progress_1000 % 10);
-	assert(s > 0);
+	ASSERT(s > 0);
 	buf[s - 1] = '\0';
 	return buf;
 }
@@ -573,13 +572,12 @@ static int cmd_replace_cancel(const struct cmd_struct *cmd,
 	ret = ioctl(fd, BTRFS_IOC_DEV_REPLACE, &args);
 	close_file_or_dir(fd, dirstream);
 	if (ret < 0) {
-		fprintf(stderr, "ERROR: ioctl(DEV_REPLACE_CANCEL) failed on \"%s\": %m",
-			path);
+		error("ioctl(DEV_REPLACE_CANCEL) failed on \"%s\": %m", path);
 		if (args.result != BTRFS_IOCTL_DEV_REPLACE_RESULT_NO_RESULT)
-			fprintf(stderr, ", %s\n",
+			pr_stderr(LOG_DEFAULT, ", %s\n",
 				replace_dev_result2string(args.result));
 		else
-			fprintf(stderr, "\n");
+			pr_stderr(LOG_DEFAULT, "\n");
 		return 1;
 	}
 	if (args.result == BTRFS_IOCTL_DEV_REPLACE_RESULT_NOT_STARTED) {
