@@ -17,35 +17,29 @@
  */
 
 #include "kerncompat.h"
-
+#include <sys/ioctl.h>
 #include <unistd.h>
-#include <stdint.h>
-#include <dirent.h>
 #include <fcntl.h>
 #include <pthread.h>
-#include <math.h>
-#include <sys/stat.h>
-#include <sys/types.h>
-#include <sys/ioctl.h>
-#include <libgen.h>
-#include <mntent.h>
-#include <assert.h>
 #include <getopt.h>
-#include <uuid/uuid.h>
 #include <limits.h>
-
-#include "kernel-shared/ctree.h"
-#include "ioctl.h"
-#include "cmds/commands.h"
-#include "kernel-lib/list.h"
+#include <errno.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include "kernel-lib/sizes.h"
 #include "common/utils.h"
-
-#include "kernel-shared/send.h"
 #include "common/send-utils.h"
 #include "common/help.h"
 #include "common/path-utils.h"
+#include "common/string-utils.h"
+#include "common/messages.h"
+#include "cmds/commands.h"
+#include "ioctl.h"
 
-#define SEND_BUFFER_SIZE	SZ_64K
+#define BTRFS_SEND_BUF_SIZE_V1	(SZ_64K)
+#define BTRFS_MAX_COMPRESSED	(SZ_128K)
+#define BTRFS_SEND_BUF_SIZE_V2	(SZ_16K + BTRFS_MAX_COMPRESSED)
 
 
 struct btrfs_send {
@@ -57,6 +51,8 @@ struct btrfs_send {
 	u64 clone_sources_count;
 
 	char *root_path;
+	u32 proto;
+	u32 proto_supported;
 };
 
 static int get_root_id(struct btrfs_send *sctx, const char *path, u64 *root_id)
@@ -108,6 +104,7 @@ static int find_good_parent(struct btrfs_send *sctx, u64 root_id, u64 *found)
 			ret = -ENOENT;
 		else
 			ret = PTR_ERR(parent);
+		parent = NULL;
 		goto out;
 	}
 
@@ -142,6 +139,7 @@ static int find_good_parent(struct btrfs_send *sctx, u64 root_id, u64 *found)
 				ret = -ENOENT;
 			else
 				ret = PTR_ERR(parent2);
+			parent2 = NULL;
 			goto out;
 		}
 		tmp = parent2->ctransid - parent->ctransid;
@@ -206,11 +204,18 @@ static void *read_sent_data(void *arg)
 	struct btrfs_send *sctx = (struct btrfs_send*)arg;
 
 	while (1) {
+		size_t splice_buf_size = BTRFS_SEND_BUF_SIZE_V1;
 		ssize_t sbytes;
 
+		if(sctx->proto > 1){
+			splice_buf_size = BTRFS_SEND_BUF_SIZE_V2;
+
+			/* Try to change pipe buffer size */
+			fcntl(sctx->dump_fd, F_SETPIPE_SZ, splice_buf_size);
+		}
 		/* Source is a pipe, output is either file or stdout */
 		sbytes = splice(sctx->send_fd, NULL, sctx->dump_fd,
-				NULL, SEND_BUFFER_SIZE, SPLICE_F_MORE);
+				NULL, splice_buf_size, SPLICE_F_MORE);
 		if (sbytes < 0) {
 			ret = -errno;
 			error("failed to read stream from kernel: %m");
@@ -257,6 +262,19 @@ static int do_send(struct btrfs_send *send, u64 parent_root_id,
 	memset(&io_send, 0, sizeof(io_send));
 	io_send.send_fd = pipefd[1];
 	send->send_fd = pipefd[0];
+	io_send.flags = flags;
+
+	if (send->proto_supported > 1) {
+		/*
+		 * Versioned stream supported, requesting default or specific
+		 * number.
+		 */
+		io_send.version = send->proto;
+		io_send.flags |= BTRFS_SEND_FLAG_VERSION;
+
+		fcntl(pipefd[0], F_SETPIPE_SZ, BTRFS_SEND_BUF_SIZE_V2);
+		fcntl(pipefd[1], F_SETPIPE_SZ, BTRFS_SEND_BUF_SIZE_V2);
+	}
 
 	if (!ret)
 		ret = pthread_create(&t_read, NULL, read_sent_data, send);
@@ -267,7 +285,6 @@ static int do_send(struct btrfs_send *send, u64 parent_root_id,
 		goto out;
 	}
 
-	io_send.flags = flags;
 	io_send.clone_sources = (__u64*)send->clone_sources;
 	io_send.clone_sources_count = send->clone_sources_count;
 	io_send.parent_root = parent_root_id;
@@ -280,15 +297,12 @@ static int do_send(struct btrfs_send *send, u64 parent_root_id,
 		ret = -errno;
 		error("send ioctl failed with %d: %m", ret);
 		if (ret == -EINVAL && (!is_first_subvol || !is_last_subvol))
-			fprintf(stderr,
+			pr_stderr(LOG_DEFAULT,
 				"Try upgrading your kernel or don't use -e.\n");
 		goto out;
 	}
-	if (bconf.verbose > 1)
-		fprintf(stderr, "BTRFS_IOC_SEND returned %d\n", ret);
-
-	if (bconf.verbose > 1)
-		fprintf(stderr, "joining genl thread\n");
+	pr_stderr(LOG_INFO, "BTRFS_IOC_SEND returned %d\n", ret);
+	pr_stderr(LOG_DEBUG, "joining genl thread\n");
 
 	close(pipefd[1]);
 	pipefd[1] = -1;
@@ -302,8 +316,9 @@ static int do_send(struct btrfs_send *send, u64 parent_root_id,
 	}
 	if (t_err) {
 		ret = (long int)t_err;
-		error("failed to process send stream, ret=%ld (%s)",
-				(long int)t_err, strerror(-ret));
+		errno = -ret;
+		error("failed to process send stream, ret=%ld (%m)",
+				(long int)t_err);
 		goto out;
 	}
 
@@ -419,6 +434,36 @@ static void free_send_info(struct btrfs_send *sctx)
 	sctx->root_path = NULL;
 }
 
+static u32 get_sysfs_proto_supported(void)
+{
+	int fd;
+	int ret;
+	char buf[32] = {};
+	char *end = NULL;
+	u64 version;
+
+	fd = sysfs_open_file("features/send_stream_version");
+	if (fd < 0) {
+		/*
+		 * No file is either no version support or old kernel with just
+		 * v1.
+		 */
+		return 1;
+	}
+	ret = sysfs_read_file(fd, buf, sizeof(buf));
+	close(fd);
+	if (ret <= 0)
+		return 1;
+	version = strtoull(buf, &end, 10);
+	if (version == ULLONG_MAX && errno == ERANGE)
+		return 1;
+	if (version > U32_MAX) {
+		warning("sysfs/send_stream_version too big: %llu", version);
+		version = 1;
+	}
+	return version;
+}
+
 static const char * const cmd_send_usage[] = {
 	"btrfs send [-ve] [-p <parent>] [-c <clone-src>] [-f <outfile>] <subvol> [<subvol>...]",
 	"Send the subvolume(s) to stdout.",
@@ -447,6 +492,11 @@ static const char * const cmd_send_usage[] = {
 	"                 does not contain any file data and thus cannot be used",
 	"                 to transfer changes. This mode is faster and useful to",
 	"                 show the differences in metadata.",
+	"--proto N        use protocol version N, or 0 to use the highest version",
+	"                 supported by the sending kernel (default: 1)",
+	"--compressed-data",
+	"                 send data that is compressed on the filesystem directly",
+	"                 without decompressing it",
 	"-v|--verbose     deprecated, alias for global -v option",
 	"-q|--quiet       deprecated, alias for global -q option",
 	HELPINFO_INSERT_GLOBALS,
@@ -466,12 +516,14 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 	char *snapshot_parent = NULL;
 	u64 root_id = 0;
 	u64 parent_root_id = 0;
-	int full_send = 1;
-	int new_end_cmd_semantic = 0;
+	bool full_send = true;
+	bool new_end_cmd_semantic = false;
 	u64 send_flags = 0;
+	u64 proto = 0;
 
 	memset(&send, 0, sizeof(send));
 	send.dump_fd = fileno(stdout);
+	send.proto = 1;
 	outname[0] = 0;
 
 	/*
@@ -487,11 +539,17 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 
 	optind = 0;
 	while (1) {
-		enum { GETOPT_VAL_SEND_NO_DATA = 256 };
+		enum {
+			GETOPT_VAL_SEND_NO_DATA = GETOPT_VAL_FIRST,
+			GETOPT_VAL_PROTO,
+			GETOPT_VAL_COMPRESSED_DATA,
+		};
 		static const struct option long_options[] = {
 			{ "verbose", no_argument, NULL, 'v' },
 			{ "quiet", no_argument, NULL, 'q' },
 			{ "no-data", no_argument, NULL, GETOPT_VAL_SEND_NO_DATA },
+			{ "proto", required_argument, NULL, GETOPT_VAL_PROTO },
+			{ "compressed-data", no_argument, NULL, GETOPT_VAL_COMPRESSED_DATA },
 			{ NULL, 0, NULL, 0 }
 		};
 		int c = getopt_long(argc, argv, "vqec:f:i:p:", long_options, NULL);
@@ -507,7 +565,7 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 			bconf_be_quiet();
 			break;
 		case 'e':
-			new_end_cmd_semantic = 1;
+			new_end_cmd_semantic = true;
 			break;
 		case 'c':
 			subvol = realpath(optarg, NULL);
@@ -539,7 +597,7 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 			free(subvol);
 			subvol = NULL;
 			free_send_info(&send);
-			full_send = 0;
+			full_send = false;
 			break;
 		case 'f':
 			if (arg_copy_path(outname, optarg, sizeof(outname))) {
@@ -571,7 +629,7 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 				goto out;
 			}
 
-			full_send = 0;
+			full_send = false;
 			break;
 		case 'i':
 			error("option -i was removed, use -c instead");
@@ -580,6 +638,18 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 		case GETOPT_VAL_SEND_NO_DATA:
 			send_flags |= BTRFS_SEND_FLAG_NO_FILE_DATA;
 			break;
+		case GETOPT_VAL_PROTO:
+			proto = arg_strtou64(optarg);
+			if (proto > U32_MAX) {
+				error("protocol version number too big %llu", proto);
+				ret = 1;
+				goto out;
+			}
+			send.proto = proto;
+			break;
+		case GETOPT_VAL_COMPRESSED_DATA:
+			send_flags |= BTRFS_SEND_FLAG_COMPRESSED;
+			break;
 		default:
 			usage_unknown_option(cmd, argv);
 		}
@@ -686,7 +756,36 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 
 	if ((send_flags & BTRFS_SEND_FLAG_NO_FILE_DATA) && bconf.verbose > 1)
 		if (bconf.verbose > 1)
-			fprintf(stderr, "Mode NO_FILE_DATA enabled\n");
+			pr_stderr(LOG_DEFAULT, "Mode NO_FILE_DATA enabled\n");
+	send.proto_supported = get_sysfs_proto_supported();
+	if (send.proto_supported == 1) {
+		if (send.proto > send.proto_supported) {
+			error("requested version %u but kernel supports only %u",
+			      send.proto, send.proto_supported);
+			ret = -EPROTO;
+			goto out;
+		}
+	}
+	if (send_flags & BTRFS_SEND_FLAG_COMPRESSED) {
+		/*
+		 * If no protocol version was explicitly requested, then
+		 * --compressed-data implies --proto 2.
+		 */
+		if (send.proto == 1 && !proto)
+			send.proto = 2;
+
+		if (send.proto == 1) {
+			error("--compressed-data requires protocol version >= 2 (requested 1)");
+			ret = -EINVAL;
+			goto out;
+		} else if (send.proto == 0 && send.proto_supported < 2) {
+			error("kernel does not support --compressed-data");
+			ret = -EINVAL;
+			goto out;
+		}
+	}
+	pr_stderr(LOG_INFO, "Protocol version requested: %u (supported %u)\n",
+		send.proto, send.proto_supported);
 
 	for (i = optind; i < argc; i++) {
 		int is_first_subvol;
@@ -695,8 +794,7 @@ static int cmd_send(const struct cmd_struct *cmd, int argc, char **argv)
 		free(subvol);
 		subvol = argv[i];
 
-		if (bconf.verbose > BTRFS_BCONF_QUIET)
-			fprintf(stderr, "At subvol %s\n", subvol);
+		pr_stderr(LOG_DEFAULT, "At subvol %s\n", subvol);
 
 		subvol = realpath(subvol, NULL);
 		if (!subvol) {
